import re

from pyparsing import *

# from acclimatise.flag_parser.elements import cli_id, any_flag, long_flag, short_flag, flag_with
from acclimatise.flag_parser.elements import (
    arg,
    argument_body_chars,
    customIndentedBlock,
    delimited_body_chars,
    element_body_chars,
    element_start_chars,
    flag_with_arg,
    repeated_segment,
)
from acclimatise.model import (
    Command,
    EmptyFlagArg,
    Flag,
    FlagSynonym,
    Positional,
    SimpleFlagArg,
)
from acclimatise.usage_parser.model import UsageElement


def delimited_item(open, el, close):
    def action(s, loc, toks):
        return toks[1:-1]

    return (open + el + close).setParseAction(action)


usage_element = Forward()
element_char = arg.copy()  # Word(initChars=element_start_chars, bodyChars=)

mandatory_element = (
    element_char.copy()
    .setParseAction(
        lambda s, loc, toks: UsageElement(
            text=toks[0],
        )
    )
    .setName("MandatoryElement")
)
"""
A mandatory element in the command-line invocation. Might be a variable or a constant
"""

variable_element = (
    delimited_item(
        "<", Word(initChars=element_start_chars, bodyChars=delimited_body_chars), ">"
    )
    .setParseAction(lambda s, loc, toks: UsageElement(text=toks[1], variable=True))
    .setName("VariableElement")
)
"""
Any element inside angle brackets is a variable, meaning you are supposed to provide your own value for it.
However, some usage formats show variables without the angle brackets
"""


def visit_optional_section(s, loc, toks):
    inner = toks[1:-1]
    for tok in inner:
        tok.optional = True
    return inner


optional_section = (
    delimited_item("[", OneOrMore(usage_element), "]")
    .setParseAction(visit_optional_section)
    .setName("OptionalSection")
)
"""
Anything can be nested within square brackets, which indicates that everything there is optional
"""

# flag_arg = Or([
#     variable_element,
#     element_char
# ])
"""
The argument after a flag, e.g. in "-b <bamlist.fofn>" this would be everything after "-b"
"""

# short_flag_name = Char(alphas)
"""
The single character for a short flag, e.g. "n" for a "-n" flag
"""

# short_flag = (
#         '-' + short_flag_name + White() + Optional(flag_arg)
# ).setParseAction(
#     lambda s, loc, toks:
#     Flag.from_synonyms([FlagSynonym(
#         name=toks[0] + toks[1],
#         argtype=SimpleFlagArg(toks[3]) if toks[3] else EmptyFlagArg()
#     )], description=None)
# )
"""
The usage can contain a flag with its argument
"""

# long_flag = (
#         '--' + element_char + White() + Optional(flag_arg)
# ).setParseAction(lambda s, loc, toks: Flag.from_synonyms([FlagSynonym(
#     name=toks[1],
#     argtype=SimpleFlagArg(toks[3]) if toks[3] else EmptyFlagArg()
# )]))
"""
The usage can contain a flag with its argument
"""


def visit_short_flag_list(s, loc, toks):
    return [
        Flag.from_synonyms(
            [FlagSynonym(name="-" + flag, argtype=EmptyFlagArg())], description=None
        )
        for flag in toks[1:]
    ]


# short_flag_list = ('-' + short_flag_name + OneOrMore(short_flag_name)).setParseAction(
#     visit_short_flag_list).leaveWhitespace()
"""
Used to illustrate where a list of short flags could be used, e.g. -nurlf indicates -n or -u etc
"""


def visit_list_element(s, loc, toks):
    # Pick the last element if there is one, otherwise use the first element
    # This gives us a better name like 'inN.bam' instead of 'in2.bam'
    els = [tok for tok in toks if isinstance(tok, (UsageElement, Flag))]
    for el in els:
        el.repeatable = True
    return els[-1]


options_placeholder = (
    Regex("options?", flags=re.IGNORECASE).suppress().setName("OptionsPlaceholder")
)

list_element = (
    (
        OneOrMore(options_placeholder ^ mandatory_element ^ variable_element)
        + Literal(".")[2, 3]
        + Optional(options_placeholder ^ mandatory_element ^ variable_element)
    )
    .setParseAction(visit_list_element)
    .setName("list_element")
)
"""
When one or more arguments are allowed, e.g. "<in2.bam> ... <inN.bam>"
"""

usage_flag = (
    And([flag_with_arg])
    .setParseAction(lambda s, loc, toks: Flag.from_synonyms(toks, description=""))
    .setName("usage_flag")
)


usage_element <<= Or(
    [
        optional_section,
        list_element,
        # short_flag_list,
        usage_flag,
        variable_element,
        options_placeholder,
        mandatory_element,
    ]
).setName("usage_element")

stack = [1]


def visit_usage(s, loc, toks):
    # Fix up stack inconsistencies
    while len(stack) > 1:
        stack.pop()

    return toks[0][0]


usage_example = OneOrMore(usage_element, stopOn=LineEnd())
"""
Each usage example is a single line of text, e.g. 

  shell [options] -e string
"""

usage = (
    LineStart()
    + Regex("usage:", flags=re.IGNORECASE).suppress()
    + OneOrMore(usage_example)
)  # .setParseAction(visit_usage).setDebug()
"""
Each usage block can have one or more lines of different usage. e.g.

Usage:
  shell [options] -e string
    execute string in V8
  shell [options] file1 file2 ... filek
    run JavaScript scripts in file1, file2, ..., filek
"""


# usage = Regex('usage:', flags=re.IGNORECASE).suppress() + delimitedList(usage_element, delim=Or([' ', '\n']))
# indentedBlock(
#     delimitedList(usage_element, delim=' '),
#     indentStack=stack,
#     indent=True
# )



import itertools
import unicodedata
from collections import Counter, defaultdict
from difflib import ndiff, unified_diff
from itertools import groupby
from typing import Generator, Iterable, List, Optional, Set, Tuple

import regex as re
from num2words import num2words
from spacy.tokens import Token
from word2number import w2n

from acclimatise.nlp import nlp, wordsegment


class NameGenerationError(Exception):
    def __init__(self, message: str):
        self.message = message


def useless_name(name: List[str]):
    """
    Returns true if this name (sequence of strings) shouldn't be used as a variable name because it's too short and
    uninformative. This includes an entirely numeric name, which is almost never what you want
    """
    joined = "".join(name)
    if len(name) < 1 or len(joined) <= 1 or joined.isnumeric():
        return True

    # Numeric names are not useful
    try:
        if all([w2n.word_to_num(tok) is not None for tok in name]):
            return True
    except Exception:
        pass

    return False


def duplicate_keys(l: list) -> Set[int]:
    """
    Identifies the indexes of duplicates in a given list
    """
    ret = set()
    for entry, indices in groupby(
        sorted(enumerate(l), key=lambda x: x[1]), key=lambda x: x[1]
    ):
        indices = list(indices)
        if len(indices) > 1 and len("".join(entry)) > 0:
            # The keys we still need to iterate on are those with duplicate commands, but if the names are empty
            # strings, we should stop iterating and try another method
            for index in indices:
                ret.add(index[0])
    return ret


def ensure_first_alpha(text):
    return re.sub(r"\b[^[:alpha:]\s]", "", text)


def sanitize_symbols(text):
    """
    Remove any non-word symbols
    """
    return re.sub("[^[:alpha:]-_., ]", "", text)


def sanitize_token(text):
    """
    Remove any non-word symbols
    """
    return re.sub("[^[:alpha:]]", "", text)


def replace_hyphens(text):
    """
    Remove any non-word symbols
    """
    return re.sub("-", "_", text)


def should_ignore(token):
    return token.pos_ not in {"NOUN", "NUM", "PRON", "PROPN", "VERB"}


def distance_to_root(token, root):
    if token == root:
        return 0
    else:
        dist = 0
        while token != root:
            dist += 1
            token = token.head
        return dist


def token_priority(token: Token, current: List[Token], key: Set[Token] = set()) -> int:
    """
    Returns a priority (where lowest means highest priority) of tokens
    :param token: The token to prioritise
    :param current: The current list of tokens in the name
    """
    if token in key:
        # Tokens that are "keywords" are the highest priority
        return token.idx - 1000

    if token.pos_ not in {"NOUN", "NUM", "PRON", "PROPN", "VERB", "ADJ", "ADV"}:
        # Tokens that have meaningful POS are prioritised
        return 1000 + token.idx
    else:
        if token.dep_ != "HEAD" and token.head not in current:
            # If this is, say, an adjective but the noun it's attached to isn't in the list, massively penalise this
            return 500 + token.idx
        else:
            return token.idx
    # if should_ignore(token):
    #     return 100 + distance_to_root(token, root)
    # else:
    #     return distance_to_root(token, root)


def remove_delims(text):
    for delims in (("\\[", "\\]"), ("{", "}"), ("'", "'"), ('"', '"')):
        text = re.sub("{}.+{}".format(*delims), "", text)

    return text


def name_to_camel(words: Iterable[str]) -> str:
    """
    Converts a list of words into camelCase
    """
    words = list(words)
    cased = [words[0].lower()] + [segment.capitalize() for segment in words[1:]]
    return "".join(cased)


def name_to_snake(words: Iterable[str]) -> str:
    """
    Converts a list of words into snake_case
    """
    return "_".join([word.lower() for word in words])


def human_readable_translate(symbol: str):
    # First, if the symbol is in this small curated list, use that
    lookup = {".": "dot", ",": "comma", "/": "slash", "\\": "backslash", "@": "at"}
    if symbol in lookup:
        return lookup[symbol]

    # Second, if the symbol is a number, convert it to natural English
    try:
        return num2words(symbol, ordinal=False, lang="en", to="cardinal")
    except Exception:
        # If it's not a number, ignore the error
        pass

    # Finally, if neither of those worked, ask unicode what it should be called
    # This isn't ideal because unicodedata calls "/" a "SOLIDUS" and "@" it calls "commercial at", which is rather
    # awkward
    return unicodedata.name(symbol[0]).lower()


def segment_string(text: str):
    """
    Divides one larger word into segments
    :param text:
    :return:
    """

    base = text.lstrip("-")

    # Replace symbols with their unicode names
    translated = re.sub(
        "[^[:alpha:]\s\-_]", lambda match: human_readable_translate(match.group()), base
    )

    dash_tokens = re.split("[-_ ]", translated)
    segment_tokens = itertools.chain.from_iterable(
        [wordsegment.segment(w) for w in dash_tokens]
    )
    return [sanitize_token(tok) for tok in segment_tokens]


def choose_unique_name(
    options: Tuple[List[str], ...], number: int, reserved: Set[Tuple[str, ...]] = []
) -> List[str]:
    """
    Given a list of possible names for each flag, choose the first one that is not too short
    """
    # First try all of them, being picky
    for option in options:
        if tuple(option) in reserved:
            # If name generation produces a reserved keyword, skip it
            continue
        if not useless_name(option):
            return option

    # Second, try all of them choosing anything that isn't empty
    for option in options:
        if tuple(option) in reserved:
            # If name generation produces a reserved keyword and we've already tried everything else, add "var" to it as a prefix
            return ["var", *option]
        if len(option) == 1 and option[0] in reserved:
            continue
        if len(option) > 0:
            return option

    # If nothing else works, just use something like 'var_3' as the variable name
    return ["var", str(number)]
    # raise NameGenerationError(
    #     "No unique names available. Selecting from {}".format(
    #         ";".join([" ".join(option) for option in options])
    #     )
    # )


def generate_names_segment(
    descriptions: List[str], initial_length: int = 3, max_length: int = 5
) -> List[List[str]]:
    """
    Given a list of short strings, one for each variable, we segment each string. Any duplicate names are set to an
    empty list as they are unusable
    """
    segmented = [segment_string(desc) for desc in descriptions]
    for dup in duplicate_keys(segmented):
        segmented[dup] = []
    return segmented

    # todo = set(range(len(descriptions)))
    # segmented = [segment_string(desc) for desc in descriptions]
    # ret = [[]] * len(descriptions)
    # empty = set()
    #
    # # Increase the maximum length each iteration
    # for length in range(initial_length, max_length + 1):
    #
    #     # If we run out of iterations and still haven't converged, these variables names become empty
    #     if length >= max_length:
    #         for j in todo:
    #             ret[j] = []
    #         return ret
    #
    #     # Update the outputs using each generator
    #     for j in todo:
    #         # If this token is still not unique, but it has no more characters left, it's a lost cause
    #         if length > len(segmented[j]):
    #             empty.add(j)
    #             ret[j] = []
    #         ret[j] = segmented[j][:length]
    #
    #     # Clean up old generators
    #     todo = duplicate_keys(ret) - empty
    #
    #     return ret


def intersection_indices(
    l: List[List[str]], reserved: Set[Tuple[str, ...]]
) -> Set[int]:
    """
    Returns a set of indices from ``l`` that occur in the set of reserved keywords
    """
    ret = set()
    for i, words in enumerate(l):
        if tuple(words) in reserved:
            ret.add(i)
    return ret


def find_key_diff_words(
    descriptions: List[List[Token]], after_index: int = 5
) -> List[Set[Token]]:
    """
    Returns a list of words for each description that uniquely identifies each
    :param after_index: The index in the string after which to consider key words. Because we always include the first
    few words in generated names, finding diffs at these positions isn't helpful
    """
    desc_strings = [[str(token) for token in desc] for desc in descriptions]
    ret = [set() for d in descriptions]
    for i, description_a in enumerate(descriptions):
        for j, description_b in enumerate(descriptions):
            if i == j:
                # Don't compare a word to itself
                continue

            # Keep iterating until we find the first diff, then add the diff words to a list of key words
            for k, diff in enumerate(ndiff(desc_strings[i], desc_strings[j])):
                if diff[0] in ["+", "-"]:
                    if k > after_index:
                        # If this token differs, add the difference
                        if len(description_a) > k:
                            ret[i].add(description_a[k])
                        if len(description_b) > k:
                            ret[j].add(description_b[k])

                    # Always break once we find a diff
                    break

    return ret


def preprocess(text: str) -> List[Token]:
    """
    Pre-process some text, remove and unnecessary tokens, and return the Spacy data structure
    """
    # Remove all delimited text
    sanitized = ensure_first_alpha(
        sanitize_symbols(remove_delims(replace_hyphens(text)))
    )

    # Parse the sentence
    doc = nlp(sanitized)

    candidates = [tok for sent in doc.sents for tok in sent]
    if len(candidates) == 0:
        return []
    else:
        return candidates


def generate_names_nlp(
    descriptions: List[str],
    initial_length: int = 3,
    max_length: int = 5,
    reserved: Set[Tuple[str, ...]] = set(),
) -> List[List[str]]:
    """
    Given a list of flag descriptions, iterates until it generates a set of unique names for each flag
    :param initial_length: The minimum/starting length for each variable name
    :param max_length: The maximum length variables can have before it will fail
    :param reserved: Keywords that are not permitted to be used as the entire name
    """
    #: A set of indices of flags that still need to be named
    todo = set(range(len(descriptions)))

    # Preprocessing
    processed = list(map(preprocess, descriptions))
    diffs = find_key_diff_words(processed)
    generators = [
        generate_name(proc, initial_length=initial_length, key_words=keywords)
        for proc, keywords in zip(processed, diffs)
    ]
    #: A list of flag names, in the same order as the input list
    ret = [[]] * len(descriptions)
    #: A set of generators that are exhausted, meaning we can't iterate on them
    empty = set()

    # Iterate until everything is done, at most 5 more times
    for i in range(max_length + 1):
        if i >= max_length:
            # If we run out of iterations and still haven't converged, these variables names become empty
            for j in todo:
                ret[j] = []
            return ret
            # raise Exception("Variable names failed to converge")

        # Update the outputs using each generator
        for j in todo:
            try:
                # If we are able to add more characters, do so
                name = next(generators[j])
                ret[j] = name
            except StopIteration:
                # If we can't, exclude it from further calculations
                empty.add(j)

        # Clean up old generators
        new_todo = duplicate_keys(ret) | intersection_indices(ret, reserved) - empty
        for j in todo - new_todo:
            generators[j] = None
        todo = new_todo

        # Finish once every variable name is unique
        if len(todo) == 0:
            break

    return ret


def generate_name(
    tokens: List[Token], initial_length: int = 3, key_words: Set[Token] = set()
) -> Generator[List[str], None, None]:
    """
    Given one or more sentences, attempt to parse out a concise (2-4 word) variable name. This is a generator, and each
    time you access it, it will return a slightly longer word, this is to help prevent collisions with other variable
    names
    :param tokens: One or more sentences of text that have been pre-processed
    :param initial_length: The number of words in the first output this produces
    """
    candidates = tokens
    ret = [*key_words]
    max = initial_length
    while True:

        # All tokens, sorted in order of importance
        candidates = sorted(
            candidates, key=lambda token: token_priority(token, ret), reverse=True
        )

        # Each time through the loop, we increase the number of candidates we use
        if candidates:
            ret.append(candidates.pop())

        # Iterate until we have N tokens, then yield the word. Then if the generator is called again, increase N by
        # 1 and continue

        if len(ret) >= max or len(candidates) == 0:
            max += 1
            # Now sort the tokens back into their original positions
            yield [
                sanitize_token(str(tok)).lower()
                for tok in sorted(ret, key=lambda tok: tok.i)
            ]



from typing import List, Optional

from dataclasses import dataclass

from acclimatise.model import Flag, Positional


@dataclass
class UsageElement:
    text: str
    """
    The name of this element, as defined in the usage section
    """

    optional: bool = False
    """
    Whether or not this element is required
    """

    variable: bool = False
    """
    True if this is a variable, ie you are supposed to replace this text with your own, False if this is a constant
    that you shouldn't change, e.g. the name of the application
    """

    # flag: bool = False
    """
    True if this is a flag (starts with dashes) and not a regular element
    """

    repeatable: bool = False
    """
    If this flag/argument can be used multiple times
    """


@dataclass
class UsageInstance:
    items: List[UsageElement]
    """
    The string of elements that make up a valid command invocation
    """

    description: Optional[str] = None
    """
    Description of this invocation
    """

    @property
    def positionals(self) -> List[Positional]:
        """
        Return all the positional arguments that could be derived from this instance
        """
        return [
            Positional(description="", position=i, name=el.text, optional=el.optional)
            for i, el in enumerate(self.items)
            if isinstance(el, UsageElement)
        ]

    @property
    def flags(self) -> List[Flag]:
        """
        Return all the flags that could be derived from this instance
        """
        return [el for el in self.items if isinstance(el, Flag)]



import itertools
import tempfile
from pathlib import Path

from acclimatise import explore_command

from .util import convert_validate, skip_not_installed


def test_premade_samtools(samtools_cmd):
    """
    Use a command tree that was generated beforehand, to quickly detect issues relating to the conversion of command
    trees
    """
    convert_validate(samtools_cmd, explore=True)


def test_premade_bedtools(bedtools_cmd):
    """
    Use a command tree that was generated beforehand, to quickly detect issues relating to the conversion of command
    trees
    """
    convert_validate(bedtools_cmd, explore=True)


@skip_not_installed("samtools")
@skip_not_installed("samtools.pl")
def test_explore_samtools_pl(yaml_converter):
    """
    Tests that commands with a non-standard file extension include their extension in the final output, and don't
    override another command with the same stem
    """
    samtools = explore_command(["samtools"], max_depth=0)
    samtools_pl = explore_command(["samtools.pl"], max_depth=0)
    with tempfile.TemporaryDirectory() as temp_dir:
        path = Path(temp_dir)
        filenames = set()
        for path, command in itertools.chain(
            yaml_converter.generate_tree(samtools, temp_dir),
            yaml_converter.generate_tree(samtools_pl, temp_dir),
        ):
            filenames.add(path.name)

        assert filenames == {"samtools.yml", "samtools.pl.yml"}



from abc import abstractmethod
from itertools import groupby, zip_longest
from os import PathLike
from pathlib import Path
from typing import Generator, Iterable, List, Set, TextIO, Tuple, Type

from dataclasses import dataclass

from acclimatise.model import CliArgument, Command, Flag
from acclimatise.name_generation import (
    NameGenerationError,
    choose_unique_name,
    generate_name,
    generate_names_nlp,
    generate_names_segment,
    name_to_camel,
    name_to_snake,
)


@dataclass
class NamedArgument:
    arg: CliArgument
    name: str


@dataclass
class WrapperGenerator:
    """
    Abstract base class for a class that converts a Command object into a string that defines a tool
    wrapper in a certain workflow language
    """

    cases = ["snake", "camel"]

    @classmethod
    def get_subclasses(cls) -> List[Type["WrapperGenerator"]]:
        return cls.__subclasses__()

    @classmethod
    def choose_converter(cls, typ) -> Type["WrapperGenerator"]:
        """
        Returns a converter subclass, given a converter type name
        :param type: The type of converter, e.g. 'cwl' or 'wdl'
        """
        for subclass in cls.__subclasses__():
            if subclass.format() == typ:
                return subclass

        raise Exception("Unknown format type")

    @classmethod
    @abstractmethod
    def format(cls) -> str:
        """
        Returns the output format that this generator produces as a string, e.g. "cwl"
        """
        pass

    @abstractmethod
    def save_to_string(self, cmd: Command) -> str:
        """
        Convert the command into a single string, ignoring subcommands
        """
        pass

    def save_to_file(self, cmd: Command, path: Path) -> None:
        """
        Write the command into a file
        """
        # By default we just write the string out, but subclasses can have different behaviour
        path.write_text(self.save_to_string(cmd))

    def generate_tree(
        self, cmd: Command, out_dir: PathLike
    ) -> Generator[Tuple[Path, Command], None, None]:
        out_dir = Path(out_dir)
        for cmd in cmd.command_tree():
            path = out_dir / (cmd.as_filename + self.suffix)
            try:
                self.save_to_file(cmd, path)
            except NameGenerationError as e:
                raise NameGenerationError(
                    'Name generation error for command "{}". {}'.format(
                        " ".join(cmd.command), e.message
                    )
                )
            yield path, cmd

    @property
    def reserved(self) -> Set[Tuple[str, ...]]:
        """
        A list of reserved keywords for this language
        """
        return set()

    @property
    @abstractmethod
    def suffix(self) -> str:
        """
        Returns a suffix for files generated using this converter
        """

    def words_to_name(self, words: Iterable[str]):
        """
        Converts a list of tokens, such as ["a", "variable", "name"] to a language-appropriate name, such as
        "aVariableName"
        """
        if self.case == "snake":
            return name_to_snake(words)
        elif self.case == "camel":
            return name_to_camel(words)

    def choose_variable_names(
        self, flags: List[CliArgument], length: int = 3
    ) -> List[NamedArgument]:
        """
        Choose names for a list of flags. This needs to be done in one go because there is a risk of duplicate
        variable names otherwise
        :param length: See :py:func:`acclimatise.name_generation.generate_name`
        """
        options = list(
            zip_longest(
                generate_names_segment([flag.full_name() for flag in flags]),
                generate_names_nlp(
                    [flag.description for flag in flags], reserved=self.reserved
                ),
                [flag.argument_name() for flag in flags if isinstance(flag, Flag)],
                fillvalue=[],
            )
        )

        return [
            NamedArgument(
                arg=flag,
                name=self.words_to_name(
                    choose_unique_name(flag_options, reserved=self.reserved, number=i)
                ),
            )
            for i, (flag, flag_options) in enumerate(zip(flags, options))
        ]

    case: str = "snake"
    """
    Which case to use for variable names
    """

    generate_names: bool = True
    """
    Rather than using the long flag to generate the argument name, generate them automatically using the
    flag description. Generally helpful if there are no long flags, only short flags.
    """

    ignore_positionals: bool = False
    """
    Don't include positional arguments, for example because the help formatting has some
    misleading sections that look like positional arguments
    """



import pytest

from acclimatise.cli_types import (
    CliBoolean,
    CliDir,
    CliFile,
    CliFloat,
    CliInteger,
    CliString,
)
from acclimatise.model import infer_type


@pytest.mark.parametrize(
    "string,typ",
    [
        ("", None),
        ("int", CliInteger()),
        ("size", CliInteger()),
        ("length", CliInteger()),
        ("max", CliInteger()),
        ("min", CliInteger()),
        ("str", CliString()),
        ("float", CliFloat()),
        ("decimal", CliFloat()),
        ("bool", CliBoolean()),
        ("file", CliFile()),
        ("path", CliFile()),
        ("input file", CliFile(output=False)),
        ("output file", CliFile(output=True)),
        ("folder", CliDir()),
        ("directory", CliDir()),
        ("output directory", CliDir(output=True)),
        ("blah 23 blub", CliInteger()),
        ("nonsense 23.42", CliFloat()),
        (".42 gibberish", CliFloat()),
        ("1E-5", CliFloat()),
    ],
)
def test_type_inference(string, typ):
    inferred_type = infer_type(string)
    assert inferred_type == typ



from typing import List

from pyparsing import *


class IndentCheckpoint(ParseElementEnhance):
    """
    This is a wrapper element that simply rolls back changes in the indent stack whenever the contained element
    fails to match. This ensures the stack remains accurate
    """

    def __init__(self, expr: ParserElement, indent_stack: List[int]):
        super().__init__(expr)
        # self.expr = expr
        self.stack = indent_stack

    def parseImpl(self, instring, loc, doActions=True):
        # Backup the stack whenever we reach this element during the parse
        backup_stack = self.stack[:]
        try:
            return self.expr._parse(instring, loc, doActions, callPreParse=False)
        except ParseException as e:
            # On a parse failure, reset the stack
            self.stack[:] = backup_stack
            raise e

    def __str__(self):
        if hasattr(self, "name"):
            return self.name

        if self.strRepr is None:
            self.strRepr = "Indented[" + str(self.expr) + "]"

        return self.strRepr


class IndentParserMixin:
    """
    A mixin that maintains an indent stack, and utility methods for them
    """

    def __init__(self):
        self.stack = [1]

    def pop_indent(self):
        def check_indent(s, l, t):
            self.stack.pop()

        return (Empty() + Empty()).setParseAction(check_indent).setName("Pop")

    def push_indent(self):
        def check_indent(s, l, t):
            curCol = col(l, s)
            self.stack.append(curCol)

        return (Empty() + Empty()).setParseAction(check_indent).setName("Push")

    def peer_indent(self, allow_greater=False):
        """
        :param allow_greater: Allow greater indent than the previous indentation, but don't add it to the stack
        """

        def check_peer_indent(s, l, t):
            if l >= len(s):
                return
            curCol = col(l, s)
            if allow_greater and curCol >= self.stack[-1]:
                return
            elif curCol == self.stack[-1]:
                return
            else:
                if curCol > self.stack[-1]:
                    raise ParseException(s, l, "illegal nesting")
                raise ParseException(s, l, "not a peer entry")

        return Empty().setParseAction(check_peer_indent).setName("Peer")

    def indent(self, update=True):
        """
        :param update: If true, update the stack, otherwise simply check for an indent
        """

        def check_sub_indent(s, l, t):
            curCol = col(l, s)
            if curCol > self.stack[-1]:
                if update:
                    self.stack.append(curCol)
            else:
                raise ParseException(s, l, "not a subentry")

        return (Empty() + Empty().setParseAction(check_sub_indent)).setName("Indent")

    def dedent(self, precise=True):
        def check_dedent(s, l, t):
            if l >= len(s):
                return
            curCol = col(l, s)
            if precise and self.stack and curCol not in self.stack:
                raise ParseException(s, l, "not an unindent")
            if curCol < self.stack[-1]:
                self.stack.pop()

        return Empty().setParseAction(check_dedent).setName("Unindent")


__all__ = [IndentCheckpoint, IndentParserMixin]



from io import StringIO
from os import PathLike
from pathlib import Path
from typing import Generator, List

from dataclasses import dataclass

from acclimatise.converter import WrapperGenerator
from acclimatise.model import Command
from acclimatise.yaml import yaml


@dataclass
class YmlGenerator(WrapperGenerator):
    """
    Internal YML format
    """

    @property
    def suffix(self) -> str:
        return ".yml"

    def save_to_file(self, cmd: Command, path: Path) -> None:
        with path.open("w") as fp:
            yaml.dump(cmd, fp)

    def save_to_string(self, cmd: Command) -> str:
        buffer = StringIO()
        yaml.dump(cmd, buffer)
        return buffer.getvalue()

    @classmethod
    def format(cls) -> str:
        return "yml"



from pathlib import Path
from typing import List

from pyparsing import *

from acclimatise.flag_parser.elements import customIndentedBlock, description_line
from acclimatise.parser import IndentCheckpoint, IndentParserMixin
from acclimatise.usage_parser.elements import usage_example
from acclimatise.usage_parser.model import UsageElement, UsageInstance

from .elements import *


def normalise_cline(tokens):
    """
    Normalise a command line string, such as ["dotnet", "Pisces.dll"], converting it to ["dotnet", "pisces"]
    :param tokens:
    :return:
    """
    return [Path(el.lower()).stem for el in tokens]


class UsageParser(IndentParserMixin):
    def __init__(self):
        super().__init__()

        def visit_description_block(s, loc, toks):
            return "\n".join(toks)

        self.description_block = IndentCheckpoint(
            self.indent()
            + (self.peer_indent(allow_greater=True) + description_line)[1, ...]
            + self.dedent(precise=False),
            indent_stack=self.stack,
        ).setParseAction(visit_description_block)

        def visit_single_usage(s, loc, toks):
            return [UsageInstance(items=list(toks), description=None)]

        self.single_usage = usage_example.copy().setParseAction(visit_single_usage)

        def visit_described_usage(s, loc, toks):
            if len(toks) > 0 and isinstance(toks[-1], str):
                description = toks[-1]
            else:
                description = None

            return UsageInstance(items=list(toks[:-1]), description=description)

        self.described_usage = (
            usage_example + Optional(self.description_block)
        ).setParseAction(visit_described_usage)

        def visit_multi_usage(s, loc, toks):
            return list(toks)

        self.multi_usage = (
            LineEnd().suppress()
            + (
                IndentCheckpoint(
                    # This indent ensures that every usage example is somewhat indented (more than column 1, at least),
                    # and also sets the baseline from which the description block is measured
                    self.indent() + self.described_usage
                    # The pop here doesn't check that we have dedented, but rather it just resets the indentation so that
                    # a new usage block can have a different indentation
                    + self.pop_indent(),
                    indent_stack=self.stack,
                )
            )[1, ...]
        ).setParseAction(visit_multi_usage)

        self.usage = (
            LineStart()
            + Regex("usage:", flags=re.IGNORECASE).suppress()
            + Optional(self.multi_usage | self.single_usage)
        ).setWhitespaceChars(
            "\t "
        )  # .setParseAction(visit_usage).setDebug()

    def parse_usage(self, cmd: List[str], usage: str, debug: bool = False) -> Command:
        # return self.usage.searchString(usage)
        usage_blocks = self.usage.setDebug(debug).searchString(usage)
        if not usage_blocks:
            # If we had no results, return an empty command
            return Command(command=cmd)

        instances = []
        all_positionals = []
        all_flags = []
        for block in usage_blocks:
            for instance in block:

                positional = [
                    tok for tok in instance.items if isinstance(tok, UsageElement)
                ]
                flags = [tok for tok in instance.items if isinstance(tok, Flag)]

                # Remove an "options" argument which is just a proxy for other flags
                # positional = [pos for pos in positional if pos.text.lower() != "options"]
                # The usage often starts with a re-iteration of the command name itself. Remove this if present
                for i in range(len(positional)):
                    # For each positional argument, if the entire cmd string is present, slice away this and everything before it
                    end = i + len(cmd)
                    if end <= len(positional) and normalise_cline(
                        [pos.text for pos in positional[i:end]]
                    ) == normalise_cline(cmd):
                        positional = positional[end:]

                if not any([tok for tok in positional if tok.variable]):
                    # If the usage didn't explicitly mark anything as a variable using < > brackets, we have to assume that
                    # everything other than flags are positional elements
                    for element in positional:
                        element.variable = True

                instances.append(instance)
                # Convert these UsageElements into Positionals
                all_positionals += [
                    Positional(
                        description="", position=i, name=el.text, optional=el.optional
                    )
                    for i, el in enumerate(positional)
                ]
                all_flags += flags

        return Command(
            command=cmd,
            positional=Positional.deduplicate(all_positionals),
            named=Flag.deduplicate(all_flags),
        )



from itertools import groupby
from operator import attrgetter

import regex

from acclimatise.flag_parser.elements import *
from acclimatise.parser import IndentCheckpoint, IndentParserMixin


def pick(*args):
    def action(s, loc, toks):
        if len(args) == 1:
            return toks[args[0]]
        else:
            return [toks[arg] for arg in args]

    return action


def unique_by(
    iterable: typing.Iterable,
    by: typing.Callable[[typing.Any], typing.Hashable],
    keep: str = "first",
) -> typing.List:
    """
    Removes duplicates from an iterable, by grouping by the result of the ``by`` argument, and then
    :param iterable: An iterable to make unique
    :param by: A function that returns a value for each element in ``iterable``
    :param keep: Either "first", "last", or "none". "none" means "delete all objects that are duplicates"
    """
    groups = groupby(iterable, by)
    ret = []
    for name, group in groups:
        group = list(group)
        if keep == "first":
            ret.append(group[0])
        elif keep == "last":
            ret.append(group[-1])
        elif keep == "none":
            if len(group) == 1:
                ret.append(group[0])
    return ret


#: Used to find normal word boundaries
single_space = regex.compile(r"\b \b")

#: Used to find unusual word boundaries
multi_space = regex.compile(r"\b[\s]{2,}\b")

# Used to find words that contain no alphabetical characters
non_alpha = regex.compile(r"^[^[:alpha:]]+$")


# The reason we have a parser class here instead of just a function is so that we can store the parser state, in
# particular the indentation stack. Without this, we would have to use a global stack which would be even more
# worrying
class CliParser(IndentParserMixin):
    def parse_command(self, cmd, name) -> Command:
        all_flags = list(itertools.chain.from_iterable(self.flags.searchString(cmd)))
        # If flags aren't unique, they likely aren't real flags
        named = unique_by(
            [flag for flag in all_flags if isinstance(flag, Flag)],
            by=attrgetter("longest_synonym"),
            keep="none",
        )
        positional = unique_by(
            [flag for flag in all_flags if isinstance(flag, Positional)],
            by=attrgetter("name"),
            keep="none",
        )
        return Command(command=name, positional=positional, named=named)

    def __init__(self):
        super().__init__()

        def parse_description(s, lok, toks):
            text = "".join(toks)
            if len(text.strip()) == 0:
                return ""

            if all([non_alpha.match(word) for word in text.split()]):
                raise ParseException(
                    "This can't be a description block if all text is numeric!"
                )

            if len(multi_space.findall(text)) > len(single_space.findall(text)):
                raise ParseException(
                    "This description block has more unusual spaces than word spaces, it probably isn't a real description"
                )

            return text

        def visit_mandatory_description(s, loc, toks):
            text = toks[0].strip()
            if len(text.strip()) == 0:
                raise ParseException("A positional argument must have a description")

        self.mandatory_description = description_line.copy().setParseAction(
            visit_mandatory_description
        )

        self.flag = (
            (flag_synonyms + description_line)
            .setName("flag")
            .setParseAction(
                lambda s, loc, toks: (
                    Flag.from_synonyms(synonyms=toks[0:-1], description=toks[-1])
                )
            )
        )
        """
        The entire flag documentation, including all synonyms and one line of description
        """

        self.positional = (
            # Unlike with flags, we have to be a bit pickier about what defines a positional because it's very easy
            # for a paragraph of regular text to be parsed as a positional. So we add a minimum of 2 spaces separation
            (positional_name + White(min=2).suppress() + self.mandatory_description)
            .setName("positional")
            .setParseAction(
                lambda s, loc, toks: Positional(
                    name=toks[0], description=toks[1], position=-1
                )
            )
        )
        """
        A positional argument, e.g. a required argument that doesn't use dashes
        """

        def visit_flags(s, loc, toks):
            # Give the correct position to the positional arguments
            processed = []
            counter = 0
            # flags = toks[0]

            for flag in toks:
                if isinstance(flag, Positional):
                    flag.position = counter
                    counter += 1
                processed.append(flag)

            return processed

        def visit_colon_block(s, loc, toks):
            toks = visit_flags(s, loc, toks)
            if len(toks) < 2:
                raise ParseException(
                    "If the block starts with a colon it must match two or more tokens",
                    loc=loc,
                )
            else:
                return toks

        self.block_element = self.flag | self.positional

        def visit_description_block(s, loc, toks):
            return "\n".join(toks)

        self.description_block = IndentCheckpoint(
            self.indent()
            + (self.peer_indent(allow_greater=True) + description_line)[1, ...]
            + self.dedent(precise=False),
            indent_stack=self.stack,
        ).setParseAction(visit_description_block)
        """
        The description block is the section of indented text after a flag. e.g. in this example:
            --use_strict (enforce strict mode)
                  type: bool  default: false
        The description block is "type: bool  default: false"
        """

        self.indented_flag = IndentCheckpoint(
            # We require that each flag is indented, but we don't check for a dedent: this allows the next flag to
            # have any indentation as long as it's more indented than the top level
            self.indent() + self.block_element,
            indent_stack=self.stack,
        )
        """
        Each flag can actually be at any indentation level, but we need to update the indent stack whenever we find one,
        so that we can identify the indented description block
        """

        def visit_flag_block(s, loc, toks):
            ret: List[Flag] = []

            # The tokens are a mix of flags and lines of text. Append the text to the previous flag
            for tok in toks:
                if isinstance(tok, CliArgument):
                    ret.append(tok)
                else:
                    # Add a newline if we already have some content
                    if len(ret[-1].description) > 0:
                        ret[-1].description += "\n"
                    ret[-1].description += tok
            return ret

        self.flag_block = (
            IndentCheckpoint(
                self.indented_flag
                + (
                    # We pop the indent if parsing a new flag, since we no longer care about the previous flag
                    IndentCheckpoint(
                        self.pop_indent() + self.indented_flag, indent_stack=self.stack
                    )
                    # We don't pop the indent until after if parsing a description block, since we need to know
                    # that flag's indentation
                    | IndentCheckpoint(self.description_block, indent_stack=self.stack)
                )[...]
                + self.pop_indent(),
                indent_stack=self.stack,
            )
        ).setParseAction(visit_flag_block)
        """
        A block of flags is one or more flags, each followed by a description block. 
        The grammar is written this way so that parsing a flag is *always* prioritised over the description block, 
        preventing certain indented flags from being missed
        """

        self.colon_block = Literal(
            ":"
        ).suppress() + self.flag_block.copy().addParseAction(visit_colon_block)
        """
        When the block is introduced by a colon, we can be more lax about parsing
        """

        self.newline_block = (
            LineStart().leaveWhitespace()
            + White().suppress()
            + self.flag_block.copy().addParseAction(visit_flags)
        )
        """
        When the block is introduced by a newline, we have to be quite strict about its contents
        """

        self.unindented_flag_block = LineStart().suppress() + (
            self.flag + Optional(self.description_block)
        )[1, ...].setParseAction(visit_flag_block)
        # )  # delimitedList(self.flag, delim='\n')
        # self.unindented_flag_block.leaveWhitespace()
        self.unindented_flag_block.skipWhitespace = False
        """
        This is a list of flags that aren't indented. Because of this, we don't parse positional elements here, because
        doing so would include basically any paragraph of text
        """

        # self.colon_block.skipWhitespace = True
        # self.newline_block.skipWhitespace = True

        # A flag block can start with a colon, but then it must have 2 or more flags. If it starts with a newline it
        # only has to have one flag at least
        self.flags = (
            self.colon_block | self.newline_block | self.unindented_flag_block
        ).setName(
            "FlagList"
        )  # .leaveWhitespace()
        self.flags.skipWhitespace = False

        self.flag_section_header = Regex("(arguments|options):", flags=re.IGNORECASE)
        self.flag_section = (self.flag_section_header + self.flags).setParseAction(
            lambda s, loc, toks: toks[1:]
        )



"""
Re-usable parser elements that aren't tied to the parser object
"""
from typing import List

from pyparsing import *

from acclimatise.model import *

#: Characters that can start a CLI element, e.g. "-@"
element_start_chars = alphanums + "@"
#: Characters that can be in the middle of a CLI element, e.g. "-some-arg"
element_body_chars = element_start_chars + "-_./\\"
#: Characters that can only be used in arguments for flags e.g. "<file.fa|file.fa.gz>"
argument_body_chars = element_body_chars + "|"
#: Characters that can be in the middle of an argument that has brackets around it, e.g. "-arg <argument with space>"
delimited_body_chars = argument_body_chars + " "

NL = OneOrMore(LineEnd().setWhitespaceChars("\t ").suppress()).setName("Newline")


def customIndentedBlock(
    blockStatementExpr, indentStack, indent=True, terminal=False, lax=False
):
    """
    Modified version of the indentedBlock construct provided by pyparsing. Allows fuzzier indent boundaries
    """
    backup_stack = indentStack[:]

    def reset_stack():
        indentStack[:] = backup_stack

    def checkPeerIndent(s, l, t):
        if l >= len(s):
            return
        curCol = col(l, s)

        # A terminal block doesn't have children, so we can assume that any sub-indent is a peer
        if terminal and curCol >= indentStack[-1]:
            return

        # If we're being lax, anything that's not a full dedent is a peer
        if lax and curCol > indentStack[-2]:
            return

        # Anything that is indented more than the previous indent level counts as a peer
        if curCol < indentStack[-1] or curCol <= indentStack[-2]:
            raise ParseException(s, l, "not a peer entry")

    def checkSubIndent(s, l, t):
        curCol = col(l, s)
        if curCol > indentStack[-1]:
            indentStack.append(curCol)
        else:
            raise ParseException(s, l, "not a subentry")

    def checkUnindent(s, l, t):
        if l >= len(s):
            indentStack.pop()
            return
        curCol = col(l, s)
        if not (indentStack and curCol < indentStack[-1]):
            raise ParseException(s, l, "not an unindent")
        indentStack.pop()

    INDENT = (Empty() + Empty().setParseAction(checkSubIndent)).setName("Indent")
    PEER = Empty().setParseAction(checkPeerIndent).setName("PeerIndent")
    UNDENT = Empty().setParseAction(checkUnindent).setName("Unindent")
    if indent:
        smExpr = Group(
            Optional(NL)
            +
            # ~ FollowedBy(blockStatementExpr) +
            INDENT
            + (OneOrMore(PEER + Group(blockStatementExpr) + Optional(NL)))
            + UNDENT
        )
    else:
        smExpr = Group(
            Optional(NL) + (OneOrMore(PEER + Group(blockStatementExpr) + Optional(NL)))
        )
    smExpr.setFailAction(lambda a, b, c, d: reset_stack())
    blockStatementExpr.ignore("\\" + LineEnd())
    return smExpr.setName("IndentedBlock")


cli_id = Word(initChars=element_start_chars, bodyChars=element_body_chars)

positional_name = Word(
    initChars=element_start_chars, bodyChars=element_body_chars, min=2
)

# short_flag = originalTextFor(Literal('-') + Word(alphanums + '@', max=1))
# """A short flag has only a single dash and single character, e.g. `-m`"""
# long_flag = originalTextFor(Literal('--') + cli_id)
# """A long flag has two dashes and any amount of characters, e.g. `--max-count`"""
any_flag = (
    originalTextFor("-" + Optional("-") + cli_id).leaveWhitespace().setName("Flag")
)
"""The flag is the part with the dashes, e.g. `-m` or `--max-count`"""

flag_arg_sep = (
    Or([Literal("="), Literal(" ")]).leaveWhitespace().setName("FlagArgSeparator")
)
"""The term that separates the flag from the arguments, e.g. in `--file=FILE` it's `=`"""

arg = Word(initChars=element_start_chars, bodyChars=argument_body_chars)
"""A single argument name, e.g. `FILE`"""

optional_args = Forward().setName("OptionalArg")


def visit_optional_args(s, lok, toks):
    if len(toks) == 1:
        return OptionalFlagArg(names=[toks[0]])
    else:
        first, _, sep, second, _ = toks
        if isinstance(second, str):
            return OptionalFlagArg(names=[first, second], separator=sep)
        elif isinstance(second, OptionalFlagArg):
            return OptionalFlagArg(names=[first] + second.names, separator=sep)


optional_args <<= (
    (arg + "[" + "," + (optional_args ^ arg) + "]")
    .setParseAction(visit_optional_args)
    .setName("OptionalArgs")
)
"""
When the flag has multiple arguments, some of which are optional, e.g.
-I FLOAT[,FLOAT[,INT[,INT]]]
"""

# simple_arg = arg.copy().setParseAction(
#     lambda s, loc, toks: SimpleFlagArg(toks[0]))
simple_arg = (
    (
        Or(
            [
                Word(initChars=element_start_chars, bodyChars=element_body_chars),
                # Allow spaces in the argument name, but only if it's enclosed in angle brackets
                (
                    Literal("<").suppress()
                    + Word(
                        initChars=element_start_chars, bodyChars=delimited_body_chars
                    )
                    + Literal(">").suppress()
                ).setName("angle_delimited_arg"),
            ]
        )
    )
    .leaveWhitespace()
    .setParseAction(lambda s, loc, toks: SimpleFlagArg(toks[0]))
).setName("SimpleArg")

repeated_segment = (
    (ZeroOrMore(arg) + Literal(".")[2, 3].suppress() + Optional(arg))
    .setParseAction(lambda s, loc, toks: RepeatFlagArg(toks[-1] or toks[0]))
    .setName("RepeatedSegment")
)  # Either ".." or "..."

list_type_arg = (
    (
        (arg + repeated_segment)
        ^ (arg + Literal("[").suppress() + repeated_segment + Literal("]").suppress())
    )
    .setParseAction(lambda s, loc, toks: toks[1])
    .setName("repeated_arg")
)
"""
    When the argument is an array of values, e.g. when the help says `--samout SAMOUTS [SAMOUTS ...]` or 
    `-i FILE1 FILE2 .. FILEn`

"""

choice_type_arg = (
    nestedExpr(opener="{", closer="}", content=delimitedList(cli_id, delim=","))
    .setParseAction(lambda s, loc, toks: ChoiceFlagArg(set(toks[0])))
    .setName("ChoiceArg")
)
"""When the argument is one from a list of values, e.g. when the help says `--format {sam,bam}`"""


def noop(s, loc, toks):
    return toks


arg_expression = (
    (
        flag_arg_sep.suppress()
        + (list_type_arg | choice_type_arg | optional_args | simple_arg)
    )
    # .leaveWhitespace()
    .setParseAction(lambda s, loc, toks: toks[0])
)
arg_expression.skipWhitespace = False
"""An argument with separator, e.g. `=FILE`"""

flag_with_arg = (
    (any_flag + Optional(arg_expression))
    .setParseAction(
        lambda s, loc, toks: (
            FlagSynonym(
                name=toks[0], argtype=toks[1] if len(toks) > 1 else EmptyFlagArg()
            )
        )
    )
    .setName("FlagWithArg")
)
flag_with_arg.skipWhitespace = True
"""e.g. `--max-count=NUM`"""

# TODO: this should be smarter, accepting ' ' or '| ' or '|' etc
synonym_delim = Word(" ,|", max=2).setParseAction(noop).leaveWhitespace()
"""
The character used to separate synonyms of a flag. Depending on the help text this might be a comma, pipe or space
"""

description_sep = White(min=1).suppress()
"""
The section that separates a flag from its description. This needs to be broad enough that it will match all different
formats of help outputs but not so broad that every single word starting with a dash will be matched as a flag
"""

# block_element_prefix = LineStart().leaveWhitespace()
block_element_prefix = (
    ((LineStart().leaveWhitespace() ^ Literal(":")) + White(min=1))
    .setName("block_element_prefix")
    .leaveWhitespace()
    .suppress()
)
"""
Each element (e.g. flag) in a list of flags must either start with a colon or nothing

e.g. in this example "index" is prefixed by a colon and "mem" is prefixed by a LineStart

Command: index         index sequences in the FASTA format
         mem           BWA-MEM algorithm
"""

flag_synonyms = delimitedList(flag_with_arg, delim=synonym_delim).setName(
    "FlagSynonyms"
)
"""
When the help lists multiple synonyms for a flag, e.g:
-n, --lines=NUM
"""


# The description of the flag
# e.g. for grep's `-o, --only-matching`, this is:
# "Print only the matched (non-empty) parts of a matching line, with each such part on a separate output line."
# desc_line = originalTextFor(SkipTo(LineEnd())).setName(
#     "DescriptionLine"
# )  # .setParseAction(success))
# desc_line = originalTextFor(
#     delimitedList(Regex("[^\s]+"), delim=" ", combine=True)
# ).leaveWhitespace()


def visit_description_line(s, loc, toks):
    return toks[0].strip()


description_line = (
    SkipTo(LineEnd(), include=True)
    .setParseAction(visit_description_line)
    .setWhitespaceChars(" \t")
).setName("DescriptionLine")



"""
Tests certain groups of flags used together
"""
from acclimatise.converter import WrapperGenerator
from acclimatise.model import EmptyFlagArg, Flag, SimpleFlagArg


def test_bedtools_window_sm():
    """
    These two flags have almost the same name, and almost the same description
    """
    flags = [
        Flag(
            synonyms=["-sm"],
            description="Only report hits in B that overlap A on the _same_ strand.",
            args=EmptyFlagArg(),
        ),
        Flag(
            synonyms=["-sm"],
            description="Only report hits in B that overlap A on the _opposite_ strand.",
            args=EmptyFlagArg(),
        ),
        Flag(
            synonyms=["-c"],
            description="For each entry in A, report the number of overlaps with B.",
            args=EmptyFlagArg(),
        ),
    ]
    args = WrapperGenerator().choose_variable_names(flags)
    assert len(set([arg.name for arg in args])) == 3


def test_same_description():
    """
    Normally we ignore one-character flag names, and instead try to read their descriptions for a more informative name.
    However, if the descriptions are identical to each other, we have to fall back to the description
    """
    flags = [
        Flag(
            synonyms=["-a"],
            description="Makes the program do a certain thing",
            args=EmptyFlagArg(),
        ),
        Flag(
            synonyms=["-b"],
            description="Makes the program do a certain thing",
            args=EmptyFlagArg(),
        ),
    ]
    names = WrapperGenerator().choose_variable_names(flags)
    assert names[0].name == "a"
    assert names[1].name == "b"


def test_same_arg():
    """
    Normally we ignore one-character flag names, and instead try to read their descriptions for a more informative name.
    However, if the descriptions are identical to each other, we have to fall back to the description
    """
    flags = [
        Flag(synonyms=["-a"], description="", args=SimpleFlagArg("SomeThing")),
        Flag(synonyms=["-b"], description="", args=SimpleFlagArg("SomeThing")),
    ]
    names = WrapperGenerator().choose_variable_names(flags)
    assert names[0].name == "a"
    assert names[1].name == "b"



#!/usr/bin/env python3
from setuptools import find_packages, setup

setup(
    packages=find_packages(exclude="test"),
    install_requires=[
        "pyparsing",
        "jinja2",
        "spacy",
        "miniwdl",
        "wordsegment",
        "inflection",
        "illusional.wdlgen==0.2.10",
        "ruamel.yaml==0.16.5",
        "click",
        "cwltool",
        "cwl-utils>=0.4",
        "regex",
        "num2words",
        "word2number",
        "psutil",
        "dataclasses",
    ],
    python_requires=">=3.6",
    entry_points={"console_scripts": ["acclimatise = acclimatise.cli:main"]},
    extras_require={
        "dev": [
            "pytest",
            "pre-commit",
            "Sphinx",
            "sphinx-click",
            "pytest-timeout",
            "docker",
        ],
    },
)



import select
import socket
import time
from select import select as original_select
from typing import List, Tuple
from unittest.mock import patch

from docker.utils.socket import consume_socket_output, demux_adaptor, frames_iter

from . import Executor


def read_socket(sock, timeout: int = None) -> Tuple[bytes, bytes]:
    """
    Reads from a docker socket, and returns everything
    :param sock: Docker socket to read from
    :param timeout: Number of seconds after which we return all data collected
    :return: A tuple of stdout, stderr
    """
    start_time = time.time()
    out = [b"", b""]
    for frame in frames_iter(sock, tty=False):
        frame = demux_adaptor(*frame)

        # If we hit the timeout, return anyawy
        if time.time() >= start_time + timeout:
            return tuple(out)

        assert frame != (None, None)

        if frame[0] is not None:
            out[0] += frame[0]
        else:
            out[1] += frame[1]
    return tuple(out)


class DockerExecutor(Executor):
    """
    An executor that runs the commands on an already-running docker Container (not an Image!)
    """

    def __init__(self, container: "docker.models.containers.Container", **kwargs):
        super().__init__(**kwargs)
        self.container = container

    def execute(self, command: List[str]) -> str:
        _, sock = self.container.exec_run(
            command, stdout=True, stderr=True, demux=True, socket=True
        )
        try:
            # These are timeouts that define how long to wait while nothing is being output
            sock._sock.settimeout(self.timeout)
            with patch.object(
                select,
                "select",
                new=lambda rlist, wlist, xlist: original_select(
                    rlist, wlist, xlist, self.timeout
                ),
            ):
                stdout, stderr = read_socket(sock, timeout=self.timeout)
        except socket.timeout as e:
            return self.handle_timeout(e)

        return (stdout or stderr or b"").decode()
